//*******************************************************************
// Stepper Motor Driver Demo and Testprogram for
//
//    Allegro A3977 MicroStepping Controller
//
// This program shows how simple a driver for this controller is.
// The controller does microsteps, 8 per phase
// As the stepping power stage the E-LAB MillAxeDriver board is used
// The AVRco Stepper Driver is operated in UserPort mode.
// Please note that the steprate is doubled in this mode
// There are 3 mega32 located on the MillBase3x board which
// contain this firmware
//*******************************************************************

program StepAxe;

{$NOSHADOW}

Device = mega32;

Import SysTick, StepPort, TWInet, TickTimer;

From system import longword, Processes;
From StepPort import StepperSema;

Define
  ProcClock      = 16000000;       {Hertz}
  SysTick        = 5;              {msec}
  StackSize      = $0070, iData;
  FrameSize      = $0100, iData;
  
  Scheduler      = iData;

  TWInode        = 4;              {default address in slave Msg}
  TWIpresc       = TWI_BR400;
  TWIframe       = 40, iData;      {buffer/packet size}
  TWInetMode     = Slave;
  TWIframeBC     = 16;             {Broadcast Frame size}

  TickTimer      = Timer2;
  
  StepMinFreq    = 100;
  StepMaxFreq    = 10000;
  StepType       = UserPort;


Implementation

{$IDATA}
{--------------------------------------------------------------}
{ Type Declarations }
type
  tStepPwr    = (pwrOff, pwrLow, pwrMed, pwrHigh);

  tAxCmd     = (acIdle, acDrivePwr, acJogSpeed, acTraverse, acStartSpeed,
                acMaxSpeed, acAccel, acClrAxesPos, acPanicStop, acStopMove,
                acClearTWI, acGetPos, acSetLimPol, acSetRefPol, acRefMove,
                acVelocity);

  tMoveMode  = (mvIdle, mvJog, mvTraverse, mvFeed, mvFeedXYZ, mvReference,
                mvVelocity);

  tAxeRxRecB = record
                 axCmd   : tAxCmd;
                 bb      : byte;
               end;

  tAxeRxRecW = record
                 axCmd   : tAxCmd;
                 ww      : word;
               end;

  tAxeRxRecI = record
                 axCmd   : tAxCmd;
                 ii      : integer;
               end;

  tAxeRxRecLi = record
                  axCmd  : tAxCmd;
                  Li     : longint;
                end;

  tAxeRxRecLw = record
                 axCmd   : tAxCmd;
                 Lw      : longword;
                end;

  tAxeTxRecLi = record
                  axCmd    : tAxCmd;
                  MoveMode : tMoveMode;
                  Li       : longint;
                end;


{--------------------------------------------------------------}
{ Const Declarations }
const
  LimSw              : byte = 0;
  RefSw              : byte = 1;

{--------------------------------------------------------------}
{ Var Declarations }
{$EEPROM}
var
 Dummy                       : word;
 
structconst
 eLimSwitchPol  : boolean       = false;
 eRefSwitchPol  : boolean       = false;

{$IDATA}
VAR
  AxeAddr                    : byte;

  AxeRxRecLi[@TWIrxBuff]     : tAxeRxRecLi;
  AxeRxRecLw[@TWIrxBuff]     : tAxeRxRecLw;
  AxeRxRecW[@TWIrxBuff]      : tAxeRxRecW;
  AxeRxRecI[@TWIrxBuff]      : tAxeRxRecI;
  AxeRxRecB[@TWIrxBuff]      : tAxeRxRecB;
  
  AxeTxRecLi[@TWItxBuff]     : tAxeTxRecLi;
  
  LimSwitchPol               : boolean;
  RefSwitchPol               : boolean;

  vStepPwr                   : tStepPwr;
  vStepStartFreq             : word;            // Hz
  vStepAccValue              : word;            // Hz
  vStepEndFreq               : word;            // Hz
  vJogEndFreq                : integer;         // Hz
  vStepCountI                : longint;
  vStepCount                 : longword;
  vJogDir                    : boolean;
  vJogFreq                   : word, locked;
  vJogTimer                  : word;
  vVelocity                  : integer;

  Position                   : longint;
  MoveMode                   : tMoveMode;
  PrevMode                   : tMoveMode;

  StepOut[@PortB]            : byte;
  StepDDR[@DDRB]             : byte;
  StepStat[@PinB]            : byte;
  
  StepPulse[@StepOut, 0]     : bit;
  StepDIR[@StepOut, 1]       : bit;
  StepRefSw[@StepStat, 2]    : bit;
  StepLimSw[@StepStat, 3]    : bit;
  StepRefLo[@StepOut, 5]     : bit;
  StepRefHi[@StepOut, 4]     : bit;

  StepTravStat[@StepOut, 6]  : bit;
  StepJogStat[@StepOut, 7]   : bit;
  LimSwitchStat[@PortC, 6]   : bit;
  RefSwitchStat[@PortC, 7]   : bit;

  vStepDir[@PinD, 2]         : bit;
  
  SwitchStat                 : byte;
  bb, idx, i                 : byte;
  

{--------------------------------------------------------------}
{ functions }
UserDevice StepperIOS(cw : boolean);
begin
  StepDir:= cw;
  StepPulse:= true;
  sDelay(20);
  StepPulse:= false;
end;

Interrupt Int1;
begin
  StepDir:= not vStepDir;
  StepPulse:= true;
  sDelay(20);
  StepPulse:= false;
end;

interrupt Timer2Comp;
begin
  if vJogTimer = 0 then
    vJogTimer:= vJogFreq;
    StepDir:= vJogDir;
    StepPulse:= true;
    ASM;
      Push    _ACCBLO
      Push    _ACCBHI
    endasm;
    if vJogDir then
      inc(Position);
    else
      dec(Position);
    endif;
    ASM;
      Pop     _ACCBHI
      Pop     _ACCBLO
    endasm;
    StepPulse:= false;
  else
    //    dec(vJogTimer);
    ASM;
      LDS       _ACCCLO, StepAxe.vJogTimer
      LDS       _ACCCHI, StepAxe.vJogTimer+1
      SBIW      _ACCCLO, 1
      STS       StepAxe.vJogTimer, _ACCCLO
      STS       StepAxe.vJogTimer+1, _ACCCHI
    ENDASM;
  endif;
end;

procedure PortInit;
begin
  StepOut:= %00001100;                 // switch pullups on
  StepDDR:= %11110011;
  // set idle
  StepTravStat:= 1;
  StepJogStat:= 1;
  AxeAddr:= (PinA and $3);
  bb:= $10 or AxeAddr;
  SetTWInodeAddr(bb);
  
  PortD:= $FF;                         // Pullups on
  DDRC:=  %11000000;
  
  MCUCR:= MCUCR or $08;                // falling edge int for INT1
  GIMSK:= GIMSK or $80;                // enable INT1
end;

procedure SetStepperPwr;
begin
  case vStepPwr of
    pwrOff  : StepRefLo:= 0;
              StepRefHi:= 0;
            |
    pwrLow  : StepRefLo:= 1;
              StepRefHi:= 0;
            |
    pwrMed  : StepRefLo:= 0;
              StepRefHi:= 1;
            |
    pwrHigh : StepRefLo:= 1;
              StepRefHi:= 1;
            |
  endcase;
end;

Process RxMasterFrame(40, 80 : iData);
var
  ww         : word;
begin
  WaitSema(TWIrxSEMA);
  case AxeRxRecB.axCmd of
    acIdle       :
                 |
    acDrivePwr   : vStepPwr:= tStepPwr(AxeRxRecB.bb);
                   SetStepperPwr;
                 |
    acJogSpeed   : vJogEndFreq:= AxeRxRecI.ii;
                   ww:= word(abs(vJogEndFreq));
                   if Sign(vJogEndFreq) then
                     vJogDir:= true;
                   else
                     vJogDir:= false;
                   endif;
                   case MoveMode of
                     mvIdle : if ww > 0 then
                                vJogTimer:= 0;
                                vJogFreq:= ww;
                                MoveMode:= mvJog;
                                StepTravStat:= 0;                      // inactive
                                StepJogStat:= 1;                       // active
                                TickTimerStart;
                              else
                                TickTimerStop;
                                PrevMode:= mvIdle;
                                MoveMode:= mvIdle;
                                StepTravStat:= 0;                      // inactive
                                StepJogStat:= 0;                       // inactive
                              endif;
                            |
                     mvJog  : if ww > 0 then
                                vJogFreq:= ww;
                                StepTravStat:= 0;                      // inactive
                                StepJogStat:= 1;                       // active
                              else
                                TickTimerStop;
                                PrevMode:= MoveMode;
                                MoveMode:= mvIdle;
                                StepTravStat:= 0;                      // inactive
                                StepJogStat:= 0;                       // inactive
                              endif;
                            |
                   endcase;
                 |
    acTraverse   : if not Bit(SwitchStat, LimSw) then
                     vStepCountI:= AxeRxRecLI.Li;
                     StepCount:= longword(abs(vStepCountI));
                     if StepCount > 0 then
                       StepEndFreq:= vStepEndFreq;
                       if Sign(vStepCountI) then
                         StepDestCW;
                       else
                         StepDestCCW;
                       endif;
                       MoveMode:= mvTraverse;
                       PrevMode:= mvIdle;
                       StepTravStat:= 1;                      // active
                       StepJogStat:= 0;                       // inactive
                     endif;
                   endif;
                 |
    acRefMove    : if ((SwitchStat and $3) = 0) and (AxeRxRecI.ii <> 0) then
                     StepStartFreq:= vStepStartFreq;
                     StepEndFreq:= vStepEndFreq;
                     if Sign(AxeRxRecI.ii) then
                       StepRampCW;
                     else
                       StepRampCCW;
                     endif;
                     MoveMode:= mvReference;
                     PrevMode:= mvIdle;
                     StepTravStat:= 1;                      // active
                     StepJogStat:= 0;                       // inactive
                   endif;
                 |
    acVelocity   : if not Bit(SwitchStat, LimSw) then
                     vVelocity:= AxeRxRecI.ii;
                     if word(abs(vVelocity)) < vStepStartFreq then
                       StepRampStop;
                     elsif MoveMode = mvVelocity then
                       StepVelocity(word(abs(vVelocity)));
                     else
                       StepEndFreq:= word(abs(vVelocity));
                       if Sign(vVelocity) then
                         StepRampCW;
                       else
                         StepRampCCW;
                       endif;
                       MoveMode:= mvVelocity;
                       PrevMode:= mvIdle;
                       StepTravStat:= 1;                      // active
                       StepJogStat:= 0;                       // inactive
                     endif;
                   endif;
                 |
    acStartSpeed : vStepStartFreq:= AxeRxRecW.ww;
                   StepStartFreq:= vStepStartFreq;
                 |
    acMaxSpeed   : vStepEndFreq:= AxeRxRecW.ww;
                   StepEndFreq:= vStepEndFreq;
                 |
    acAccel      : vStepAccValue:= AxeRxRecW.ww;
                   StepAccValue:= vStepAccValue;
                 |
    acSetLimPol  : LimSwitchPol:= boolean(AxeRxRecB.bb);
                   eLimSwitchPol:= LimSwitchPol;
                 |
    acSetRefPol  : RefSwitchPol:= boolean(AxeRxRecB.bb);
                   eRefSwitchPol:= RefSwitchPol;
                 |
  endcase;
  TWIRXCLEAR;
end;

Process RxBroadcastFrame(40, 80 : iData);
begin
  WaitSema(TWIbroadcastSema);
  case TWIbroadcastCMD of
    acClrAxesPos : Position:= 0;
                 |
    acGetPos     : AxeTxRecLi.axCmd:= acGetPos;
                   AxeTxRecLi.MoveMode:= PrevMode;
                   AxeTxRecLi.Li:= Position;
                   TWItxFrame(02, byte(sizeOf(AxeTxRecLi)));
                 |
    acPanicStop  : StepPanicStop;                           // emergency stop
                   MoveMode:= mvIdle;
                   PrevMode:= MoveMode;
                   TickTimerStop;
                   StepTravStat:= 0;                        // inactive
                   StepJogStat:= 0;                         // inactive
                 |
    acStopMove   : TickTimerStop;
                   if MoveMode = mvJog then
                     PrevMode:= MoveMode;
                     MoveMode:= mvIdle;
                     StepTravStat:= 0;                      // inactive
                     StepJogStat:= 0;                       // inactive
                   elsif MoveMode = mvTraverse then
                     StepDestStop;                          // decelerate
                   elsif MoveMode in[mvReference, mvVelocity] then
                     StepRampStop;                          // decelerate
                   endif;
                 |
    acClearTWI   : TWIrxClear;
                   TWItxClear;
                 |
  endcase;
end;

(*

Process WaitStepStop(40, 80 : iData);
begin
  WaitSema(StepperSema);
  PrevMode:= MoveMode;
  MoveMode:= mvIdle;
  StepTravStat:= 0;                      // inactive
  StepJogStat:= 0;                       // inactive
end;
*)

{--------------------------------------------------------------}
{ Main Program }
{$IDATA}
begin
  PortInit;
  TWIadrMask:= $FF;
  vStepStartFreq:= 100;                      // Hz
  vStepAccValue:= 1000;                      // Hz
  vStepEndFreq:= 5000;                       // Hz
  StepStartFreq:= vStepStartFreq;
  StepAccValue:= vStepAccValue;
  StepEndFreq:= vStepEndFreq;
  TickTimerTime(50);                         // 50usec = 20kHz
  StepTravStat:= 0;                          // inactive
  StepJogStat:= 0;                           // inactive
  LimSwitchPol:= eLimSwitchPol;
  RefSwitchPol:= eRefSwitchPol;

  Priority(main_proc, 2);
  Start_Processes;

  loop
    if (MoveMode in[mvTraverse, mvReference, mvVelocity]) and (StepMode = StepStop)
    or (SemaStat(StepperSema) <> 0) then
      SetSema(StepperSema, 0);
      PrevMode:= MoveMode;
      MoveMode:= mvIdle;
      StepTravStat:= 0;                      // inactive
      StepJogStat:= 0;                       // inactive
    endif;
    
    if StepRefSw then
      if RefSwitchPol then
        RefSwitchStat:= 1;
        Incl(SwitchStat, RefSw);
      else
        RefSwitchStat:= 0;
        Excl(SwitchStat, RefSw);
      endif;
    else
      if RefSwitchPol then
        RefSwitchStat:= 0;
        Excl(SwitchStat, RefSw);
      else
        RefSwitchStat:= 1;
        Incl(SwitchStat, RefSw);
      endif;
    endif;
    if StepLimSw then
      if LimSwitchPol then
        LimSwitchStat:= 1;
        Incl(SwitchStat, LimSw);
      else
        LimSwitchStat:= 0;
        Excl(SwitchStat, LimSw);
      endif;
    else
      if LimSwitchPol then
        LimSwitchStat:= 0;
        Excl(SwitchStat, LimSw);
      else
        LimSwitchStat:= 1;
        Incl(SwitchStat, LimSw);
      endif;
    endif;

    if (MoveMode in[mvTraverse, mvReference, mvVelocity]) and Bit(SwitchStat, LimSw) then
      StepPanicStop;                           // emergency stop
      MoveMode:= mvIdle;
      PrevMode:= MoveMode;
      TickTimerStop;
      StepTravStat:= 0;                        // inactive
      StepJogStat:= 0;                         // inactive
    elsif (MoveMode = mvReference) and Bit(SwitchStat, RefSw) then
      StepPanicStop;                           // emergency stop
      MoveMode:= mvIdle;
      PrevMode:= MoveMode;
      TickTimerStop;
      StepTravStat:= 0;                        // inactive
      StepJogStat:= 0;                         // inactive
    endif;

    mDelay(10);
  endloop;
end StepAxe.


